(* format:
 *   name arg, "arg", arg ...
 *   # comment
 *)

unit SConfFile;

interface

uses SUtil;

type
  TSConfFile = class(TObject)
  private
    FFile: TSFile;
  public
    constructor Create(const FileName: string);
    destructor Destroy; override;
    function GetLine(var Name: string; var ArgCount: Integer;
      var Args: array of string): Boolean;
  end;

implementation

type
  TSetOfChar = set of Char;

function GetToken(var p: PChar; const Sep: TSetOfChar): PChar;
const
  space = [#9, ' '];
  eol = [#0, #13, #10];
  quote = ['"'];
var
  sp, q: PChar;
begin
  while p^ in space do
    Inc(p);

  Result := p;

  if p^ in quote then
  begin
    Inc(p);
    while not ((p^ in quote) or (p^ in eol)) do
      Inc(p);
  end;

  sp := p;
  while not ((p^ in Sep) or (p^ in eol)) do
    Inc(p);
  // trim tail spaces
  q := p - 1;
  while (q > sp) and (q^ in space) do
  begin
    q^ := #0;
    Dec(q);
  end;

  if p^ in Sep then
  begin
    p^ := #0;
    Inc(p);
  end else
    p^ := #0;
end;

constructor TSConfFile.Create(const FileName: string);
begin
  FFile := TSFile.Create(FileName, 'r');
end;

destructor TSConfFile.Destroy;
begin
  FFile.Free;
end;

function TSConfFile.GetLine(var Name: string; var ArgCount: Integer;
  var Args: array of string): Boolean;
var
  i: Integer;
  s: string;
  p, q: PChar;
begin
  Result := False;

  if not FFile.Opened then
    Exit;

  while True do
  begin
    Result := FFile.ReadLine(s);
    if not Result then
      Exit;

    p := PChar(s);
    Name := GetToken(p, [' ', #9]);
    if (Name = '') or (Name[1] = '#') then
      Continue;

    for i := 0 to High(Args) do
      Args[i] := '';

    ArgCount := 0;
    while (p^ <> #0) and (ArgCount <= High(Args)) do
    begin
      q := GetToken(p, [',']);
      if q[0] = '"' then
        Inc(q);
      if q[Length(q)-1] = '"' then
        q[Length(q)-1] := #0;

      Args[ArgCount] := q;
      Inc(ArgCount);
    end;
    Exit;
  end;
end;

end.
