unit NkVImage;

interface

uses Windows, SUtil, ImgFilter;

type
  TMaxBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    bmiColors: array[0..255+3] of TRGBQuad;
  end;

  TNkvImage = class(TObject)
  private
    BitmapInfo: TMaxBitmapInfo;
  public
    Loaded: Boolean;
    Width: Integer;
    Height: Integer;
    Bits: PChar;

    BitsSize: Integer;
    Bpp: Integer;
    OrgSize: Integer;
    FormatName: string;
    Comment: string;
    FilterName: string;
    FilterFileName: string;

    function ReadMem(Buf: PChar; BufSize: Cardinal; const Filter: TImgFilter): Boolean;
    procedure Draw(hDC: HDC; Left, Top, AWidth, AHeight,
      DestLeft, DestTop, DestWidth, DestHeight: Integer; ScaleFilter: Boolean);
    function Write(const FileName: string; AWidth, AHeight: Integer;
      ScaleFilter: Boolean): Boolean;
  end;

implementation

uses ImgScaler;

procedure GetImageFormat(Buf: PChar; var MimeType: string; var Name: string);
begin
  MimeType := '';
  Name := '';
  if Copy(Buf, 1, 3) = 'GIF' then
  begin
    MimeType := 'image/gif';
    Name := Copy(Buf, 1, 6);
  end
  else if Copy(Buf, 1, 2) = #$FF#$D8 then
  begin
    Inc(Buf, 6);
    if Copy(Buf, 1, 4) = 'JFIF' then
    begin
      MimeType := 'image/jpeg';
      Name := 'JPEG (JFIF)';
    end
    else if Copy(Buf, 1, 4) = 'Exif' then
    begin
      MimeType := 'image/jpeg';
      Name := 'Exif';
    end;
  end
  else if Copy(Buf, 1, 8) = #$89'PNG'#$0D#$0A#$1A#$0A then
  begin
    MimeType := 'image/png';
    Name := 'PNG';
    case Buf[28] of
      #0: Name := Name + ' - non-interlaced';
      #1: Name := Name + ' - interlaced';
    end;
    if Ord(Buf[25]) and 4 <> 0 then
      Name := Name + ' (with alpha)';
  end
  else if (Buf[0] = 'P') and (Buf[1] in ['1'..'7']) and (Buf[2] in [' ', '#', #9, #13, #10]) then
  begin
    case Buf[1] of
      '1': begin
             MimeType := 'image/x-portable-bitmap';
             Name := 'PBM image text';
           end;
      '2': begin
             MimeType := 'image/x-portable-greymap';
             Name := 'PGM image text';
           end;
      '3': begin
             MimeType := 'image/x-portable-pixmap';
             Name := 'PPM image text';
           end;
      '4': begin
             MimeType := 'image/x-portable-bitmap';
             Name := 'PBM "rawbits" image data';
           end;
      '5': begin
             MimeType := 'image/x-portable-greymap';
             Name := 'PGM "rawbits" image data';
           end;
      '6': begin
             MimeType := 'image/x-portable-pixmap';
             Name := 'PPM "rawbits" image data';
           end;
      {
      '7': begin
             MimeType := 'image/';
             Name := 'PAM image file';
           end;
      }
    end;
  end;
end;

function TNkVImage.ReadMem(Buf: PChar; BufSize: Cardinal; const Filter: TImgFilter): Boolean;
var
  size, bp: Integer;
  bmi: TMaxBitmapInfo;
  tmpbits: PChar;
  imginfo: TImgInfo;
  mimetype: string;
  fname: array[0..MAX_PATH] of Char;
begin
  Result := Filter.GetInfo(Buf, BufSize, imginfo);
  if not Result then
    Exit;

  if imginfo.Bpp in [1, 4, 8, 16, 24, 32] then
    bp := imginfo.Bpp
  else
    bp := 24;
  size := (((imginfo.Width * bp + 7) shr 3 + 3) and not 3) * imgInfo.Height;
  tmpbits := Pointer(GlobalAlloc(0, size));

  Result := Filter.Decode(Buf, BufSize, tmpbits, size, @bmi);
  if not Result then
  begin
    GlobalFree(HGLOBAL(tmpbits));
    Exit;
  end;

  // Load success
  if Bits <> nil then
    GlobalFree(HGLOBAL(Bits));
  BitmapInfo := bmi;
  Bits := tmpbits;
  BitsSize := size;
  Width := imginfo.Width;
  Height := imginfo.Height;
  Bpp := imginfo.Bpp;
  if imginfo.FormatName = '' then
    GetImageFormat(Buf, mimetype, FormatName)
  else
    FormatName := imginfo.FormatName;
  Comment := ConvLineBreak(imginfo.Comment, #13#10);
  FilterName := Filter.Name;
  if Filter.Handle = 0 then
    FilterFileName := ''
  else begin
    GetModuleFileName(Filter.Handle, fname, SIzeOf(fname));
    FilterFileName := fname;
  end;
  OrgSize := BufSize;
  Loaded := True;
end;

procedure ScaleDIB(DestBmi: PBitmapInfo; DestBits: Pointer;
  DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer;
  const SrcBmi: PBitmapInfo; SrcBits: Pointer; Filter: Boolean);
begin
  ViewTop := DestHeight - ViewTop - ViewHeight; // DIBȂ̂ŃCi[͋t

  with SrcBmi.bmiHeader do
    if Filter and (biBitCount = 24) then
    begin
      if (DestWidth > biWidth) or (biHeight div DestWidth < 2) then
        Bilinear24(SrcBits, DestBits, biWidth, biHeight, DestWidth, DestHeight,
          ViewLeft, ViewTop, ViewWidth, ViewHeight)
      else
        Average24(SrcBits, DestBits, biWidth, biHeight, DestWidth, DestHeight,
          ViewLeft, ViewTop, ViewWidth, ViewHeight);
    end else
      Nearest(SrcBits, DestBits, biWidth, biHeight, biBitCount,
        DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight);

  Move(SrcBmi^, DestBmi^, SizeOf(TMaxBitmapInfo));
  DestBmi.bmiHeader.biWidth := ViewWidth;
  DestBmi.bmiHeader.biHeight := ViewHeight;
end;

procedure TNkVImage.Draw(hDC: HDC; Left, Top, AWidth, AHeight,
  DestLeft, DestTop, DestWidth, DestHeight: Integer; ScaleFilter: Boolean);
var
  bmi: TMaxBitmapInfo;
  tmpbits: Pointer;
begin
  if (DestWidth = Width) and (DestHeight = Height) then
    SetDIBitsToDevice(hDC, Left, Top, DestWidth, DestHeight, DestLeft, 0,
      DestTop, DestHeight, Bits, PBitmapInfo(@BitmapInfo)^, DIB_RGB_COLORS)
  else begin
    //GetMem(DelphĩǗ}l[W)Xy[XfꂽubN
    //ԂꍇSetDIBitɎs\̂GlobalAllocg
    tmpbits := Pointer(GlobalAlloc(GMEM_FIXED,
      CalcPitch(AWidth, BitmapInfo.bmiHeader.biBitCount) * AHeight));
    ScaleDIB(@bmi, tmpbits, DestWidth, DestHeight, DestLeft, DestTop, AWidth, AHeight,
      @BitmapInfo, Bits, ScaleFilter);
    SetDIBitsToDevice(hDC, Left, Top, AWidth, AHeight, 0, 0, 0, AHeight, tmpbits,
      PBitmapInfo(@bmi)^, DIB_RGB_COLORS);
    GlobalFree(HGLOBAL(tmpbits));
  end;
end;

function TNkVImage.Write(const FileName: string; AWidth, AHeight: Integer;
  ScaleFilter: Boolean): Boolean;
var
  bfh: TBitmapFileHeader;
  bmi: TMaxBitmapInfo;
  bmisize, bitssize: Integer;
  tmpbits: Pointer;
begin
  bitssize := CalcPitch(AWidth, BitmapInfo.bmiHeader.biBitCount) * AHeight;
  if (AWidth = Width) and (AHeight = Height) then
  begin
    bmi := BitmapInfo;
    tmpbits := Bits;
  end else
  begin
    GetMem(tmpbits, bitssize);
    ScaleDIB(@bmi, tmpbits, AWidth, AHeight, 0, 0, AWidth, AHeight,
      @BitmapInfo, Bits, ScaleFilter);
  end;
  try
    with TSFile.Create(FileName, 'w') do
      try
        if not Opened then
        begin
          Result := False;
          Exit;
        end;

        with bmi do
        begin
          bmisize := SizeOf(bmiHeader);
          if bmiHeader.biBitCount <= 8 then
            Inc(bmisize, (1 shl bmiHeader.biBitCount) * SizeOf(TRGBQUAD));
        end;

        FillChar(bfh, SizeOf(bfh), 0);
        bfh.bfType := Byte('B') + Byte('M') shl 8;
        bfh.bfSize := SizeOf(bfh) + bmisize + bitssize;
        bfh.bfOffBits := SizeOf(bfh) + bmisize;
        // TODO: check err
        Write(bfh, SizeOf(bfh));
        Write(bmi, bmisize);
        Write(tmpbits^, bitssize);
      finally
        Free;
      end;
  finally
    if tmpbits <> Bits then
      FreeMem(tmpbits);
  end;
  Result := True;
end;

end.
